/*
 * http://www.wayforward.net/avGPS/libav/av_comm_win.c
 * Copyright(c) 2005, 2008 Wayforward LLC
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307
 * USA
 */
#define WIN32_LEAN_AND_MEAN
#include <windows.h>
#include <winsock.h>

#include <assert.h>
#include <string.h>
#include <stdlib.h>

#include "av_comm.h"
#include "av_err.h"
#include "av_str.h"

static const tchar_t SOCKET_PREFIX[] = T("socket:");
static const tchar_t SOCKET_FN[] = T("socket");
static const tchar_t GETHOSTBYNAME_FN[] = T("gethostbyname");
static const tchar_t CONNECT_FN[] = T("connect");
static const tchar_t CREATEFILE_FN[] = T("CreateFile");
static const tchar_t GETCOMMSTATE_FN[] = T("GetCommState");
static const tchar_t SETCOMMSTATE_FN[] = T("SetCommState");
static const tchar_t READFILE_FN[] = T("ReadFile");
static const tchar_t RECV_FN[] = T("recv");

static char *
tstr_to_ascii(const tchar_t *psz) {
	size_t l = tstrlen(psz) + 1;
	char *result = malloc(l);
	unsigned i;

	for (i = 0; i < l; i++)
		result[i] = (char) (psz[i] & 0x7F);

	return result;

}

/**
 * Open a comm device.  Sets baud rate, bits, parity, and stop bits.
 */
int
comm_open(const tchar_t *devicename, const struct comm_settings *pc,
	  struct comm_device *pd) {

	HANDLE hComm;

	assert(pc->baud > 0);
	assert(5 <= pc->bits && pc->bits <= 8);

	if (memcmp(devicename, SOCKET_PREFIX,
		   sizeof(SOCKET_PREFIX) - sizeof(tchar_t)) == 0) {
		static int initted = 0;

		struct sockaddr_in sin;
		char *hostname = tstr_to_ascii(devicename + 7);
		char *portstr = strchr(hostname, ':');
		int fd;

		if (!initted) {
			WSADATA wsadata;
			WSAStartup(MAKEWORD(1, 1), &wsadata);
		}

		sin.sin_family = AF_INET;

		if (portstr != NULL) {
			*portstr++ = '\0';
			sin.sin_port = htons((unsigned short) atoi(portstr));
		}
		else
			sin.sin_port = htons(2947);

		fd = socket(AF_INET, SOCK_STREAM, IPPROTO_TCP);
		if (fd < 0) {
			err_push_win32_error(SOCKET_FN, devicename + 7,
					     WSAGetLastError());
			free(hostname);
			return 0;
		}

		sin.sin_addr.s_addr = inet_addr(hostname);
		if (sin.sin_addr.s_addr == -1) {
			struct hostent *phe = gethostbyname(hostname);

			if (phe == NULL) {
				err_push_win32_error(GETHOSTBYNAME_FN, devicename + 7,
						     WSAGetLastError());
				free(hostname);
				return 0;
			}

			sin.sin_addr = *(struct in_addr *) phe->h_addr;
		}

		free(hostname);

		if (connect(fd, (struct sockaddr *) &sin, sizeof(sin)) < 0) {
			err_push_win32_error(CONNECT_FN, devicename + 7,
					     WSAGetLastError());
			return 0;
		}

		pd->hComm = (HANDLE) fd;
		pd->type = comm_socket;
		pd->running = 1;
		return 1;
	}

	hComm = CreateFile(devicename,
			   GENERIC_READ | GENERIC_WRITE,
			   0,
			   NULL,
			   OPEN_EXISTING,
			   FILE_ATTRIBUTE_NORMAL,
			   NULL);

	if (hComm == INVALID_HANDLE_VALUE) {
		err_push_last_error(CREATEFILE_FN, devicename);
		return 0;
	}

#ifdef _WIN32_WCE
	pd->hComm = hComm;
	pd->running = 1;
	pd->type = comm_tty;
	return 1;

#else
	if (GetFileType(hComm) != FILE_TYPE_CHAR) {
		pd->hComm = hComm;
		pd->type = comm_file;
		pd->running = 1;
		return 1;
	}
	else if (!GetCommState(hComm, &pd->oldState))
		err_push_last_error(GETCOMMSTATE_FN, devicename);
	else {
		DCB newState = pd->oldState;

		newState.BaudRate = pc->baud;
		newState.ByteSize = pc->bits;

		switch (pc->parity) {
			case no_parity:
				newState.Parity = NOPARITY;
				break;

			case odd_parity:
				newState.Parity = ODDPARITY;
				break;

			case even_parity:
				newState.Parity = EVENPARITY;
				break;

			case mark_parity:
				newState.Parity = MARKPARITY;
				break;

			case space_parity:
				newState.Parity = SPACEPARITY;
				break;
		}

		switch (pc->stop_bits) {
			case stop_bits_1:
				newState.StopBits = ONESTOPBIT;
				break;

			case stop_bits_1_5:
				newState.StopBits = ONE5STOPBITS;
				break;

			case stop_bits_2:
				newState.StopBits = TWOSTOPBITS;
				break;
		}

		if (!SetCommState(hComm, &newState))
			err_push_last_error(SETCOMMSTATE_FN, devicename);
		else {
			pd->hComm = hComm;
			pd->type = comm_tty;
			pd->running = 1;

			return 1;
		}
	}

	(void) CloseHandle(hComm);

	return 0;
#endif

}

#define LINE_LEN (512)

/**
 * Read lines from comm device, passing each line, including
 * terminating LF, to callback function.
 *
 * Returns 0 on error.
 * Returns 1 on EOF.
 * Returns 2 if shut down via comm_stop()
 */
int
comm_loop(struct comm_device *pd,
	  int (*callback)(void *argument, char *line, size_t len),
	  void *argument) {
	char buf[LINE_LEN + 1];
	unsigned i = 0;

	while (pd->running) {
		DWORD dwRead;

		if (pd->type == comm_socket) {
			int l;
			l = recv((int) pd->hComm, buf + i, LINE_LEN - i, 0);
			if (l < 0) {
				err_push_win32_error(RECV_FN, NULL,
						     WSAGetLastError());
				return 0;
			}

			dwRead = l;
		}
		else if (!ReadFile(pd->hComm, buf + i, LINE_LEN - i,
			      &dwRead, NULL)) {
			err_push_last_error(READFILE_FN, NULL);
			return 0;
		}

		if (dwRead == 0)
			return 1;

		i += dwRead;
		buf[i] = '\0';

		/* now consume as many lines as possible */
		i = str_iter_lines(buf, callback, argument);
	}

	return 2;
}

void
comm_close(struct comm_device *pd) {
	switch(pd->type) {
	case comm_socket:
		(void) closesocket((int) pd->hComm);
		break;

	case comm_tty:
		(void) SetCommState(pd->hComm, &pd->oldState);

		/* fall through */

	default:
		(void) CloseHandle(pd->hComm);
	}
}

/*
 * Overrides for Emacs so that we follow Linux's tabbing style.
 * Emacs will notice this stuff at the end of the file and automatically
 * adjust the settings for this buffer only.  This must remain at the
 * end of the file.
 * ----------------------------------------------------------------------
 * Local Variables:
 * c-file-style: "linux"
 * End:
 */
