/*
 * http://www.wayforward.net/avGPS/libav/av_db.c
 * Copyright(c) 2005, 2008 Wayforward LLC
 *
 * vi:set noexpandtab tabstop=8 shiftwidth=8 softtabstop=8:
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307
 * USA
 */
#include <assert.h>
#include <limits.h>

#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <math.h>

#include "zlib.h"  /* for crc32 */

#include "av_db.h"
#include "av_dir.h"
#include "av_err.h"
#include "av_mem.h"
#include "av_str.h"

enum navdb_status {
	navdb_ok,              /* success */

	/* these error code has additional OS-specific info */
	navdb_err_directory,   /* cannot open data directory */
	navdb_err_system,      /* cannot open file           */

	/* these error codes do -not- have errno info */
	navdb_no_files,        /* no files in data directory     */

	navdb_err_not_nav,     /* file is not a NAV database     */
	navdb_err_endian,      /* file has incorrect byte-order  */
	navdb_err_crc,         /* file has failed CRC-32 check   */
	navdb_err_corrupt,     /* file is a corrupt HSI database */

	navdb_err_obsolete,    /* data effective date is past    */
	navdb_err_future,      /* data effective date not yet    */

	navdb_err_version,     /* wrong database version         */
	navdb_err_missing,     /* missing required table(s)      */

	navdb_err_last
};

enum db {
	db_effective,

	db_airports,
	db_runways,
	db_navaids,
	db_fixes,
	db_obstructions,

	db_by_latitude,
	db_by_ident,

	db_routes,
	db_route_navaids,
	db_navaid_routes,

	num_dbs
};

struct nav_effective {
	struct av_datetime start, end;
};

struct nav_by_latitude {
	int ilat;
	unsigned id;
};

struct nav_by_ident {
	char ident[16];
	unsigned id;
};

struct nav_navaid_route {
	unsigned id;     /* source navaid */
	unsigned route_id;
	unsigned to_id;  /* tip navaid */
};

struct db_header {
	char tag[4];
	char name[16];
	unsigned int num_records;
	unsigned int record_length;
	unsigned int endian_check;
};

struct db_trailer {
	char tag[4];
	char name[16];
	unsigned int num_records;
	unsigned int record_length;
	unsigned int crc32;
};

static enum navdb_status
open_db(const tchar_t *fn, av_usecs now,
	struct nav_effective *p_effective_dates);

static enum navdb_status
load_db(const void *pv, size_t len, av_usecs now,
	struct nav_effective *p_effective_dates);

static const tchar_t *
navdb_strerror(enum navdb_status);

static void
nearest(void *pv, const struct nav_item *pnavitem);

static const struct nav_item *
get_pointer(unsigned id);

static void
get_ident(const struct nav_item *pnav, char buf[16], const char **psrc,
	  unsigned *plen);

static struct mapped_file s_mf;

/* pointers to individual tables in database */
static const char *s_p_dbs[num_dbs];
static const char *s_e_dbs[num_dbs];

static const char TIME_FORMAT[] = "%04d-%02d-%02dT%02d:%02d:%02dZ";

static const tchar_t ERR_DIRECTORY[] = T("cannot open directory");
static const tchar_t ERR_SYSTEM[] = T("cannot open file");
static const tchar_t NO_FILES[] = T("no files in data directory");
static const tchar_t ERR_NOT_NAV[] = T("file is not an avGPS database");
static const tchar_t ERR_ENDIAN[] = T("file is compiled for different CPU architecture");
static const tchar_t ERR_CRC[] = T("file is corrupt - CRC check failed");
static const tchar_t ERR_CORRUPT[] = T("file is corrupt - bad trailer");
static const tchar_t ERR_OBSOLETE[] = T("file is obsolete");
static const tchar_t ERR_FUTURE[] = T("file isn't effective yet");
static const tchar_t ERR_VERSION[] = T("file format mismatch");
static const tchar_t ERR_MISSING[] = T("file is missing required table");
static const tchar_t ERR_UNKNOWN[] = T("unknown navdb error");

static const struct {
	char    name[16];
	enum db db;
	size_t  record_length;
} TABLES[] = {
	{ "effective",     db_effective,     sizeof(struct nav_effective)    },
	{ "airports",      db_airports,      sizeof(struct nav_airport)      },
	{ "runways",       db_runways,       sizeof(struct nav_runway)       },
	{ "navaids",       db_navaids,       sizeof(struct nav_navaid)       },
	{ "fixes",         db_fixes,         sizeof(struct nav_fix)          },
	{ "obstructions",  db_obstructions,  sizeof(struct nav_obstruction)  },
	{ "routes",        db_routes,        sizeof(struct nav_route)        },
	{ "by_latitude",   db_by_latitude,   sizeof(struct nav_by_latitude)  },
	{ "by_ident",      db_by_ident,      sizeof(struct nav_by_ident)     },
	{ "navaid_routes", db_navaid_routes, sizeof(struct nav_navaid_route) },
};

#ifdef _WIN32
const tchar_t NAVDB_PATH[] = T("data");

#else
# ifdef __APPLE_CC__
const tchar_t NAVDB_PATH[] = T("data:~/Library/avGPS/data:/Library/avGPS/data");

# else
const tchar_t NAVDB_PATH[] = T("data:~/.avGPS/data");

# endif
#endif

#define NUM_WAYPOINTS (16)
static struct nav_fix s_waypoints[NUM_WAYPOINTS];
static unsigned s_nwaypoints, s_iwaypoint;

/**
 * Loads the navigation database, given today's date.
 *
 * Looks for the correct database using db-effective date/times.
 *
 * Returns the effective start and end dates for the database used.
 */
int
navdb_init(const tchar_t *const *path, const struct av_datetime *pdt) {
	static const tchar_t NAVDB_INIT[] = T("navdb_init");
	static const tchar_t NAVDB_NOT_FOUND[] = T("cannot find database file");
	av_usecs now, latest = 0;
	tchar_t latest_filepath[1025];

	if (pdt == NULL) {
		struct av_datetime dt_now;
		av_now(&dt_now);
		now = av_datetime_to_usecs(dt_now);
	}
	else
		now = av_datetime_to_usecs(*pdt);

	assert(sizeof(struct nav_airport) == 20);
	assert(sizeof(struct nav_runway) == 32);
	assert(sizeof(struct nav_navaid) == 32);
	assert(sizeof(struct nav_fix) == 20);
	assert(sizeof(struct nav_obstruction) == 16);
	assert(sizeof(struct nav_route) == 76);
	assert(sizeof(struct nav_by_latitude) == 8);
	assert(sizeof(struct nav_by_ident) == 20);
	assert(sizeof(struct nav_effective) == 24);
	assert(sizeof(struct db_header) == 32);
	assert(sizeof(struct db_trailer) == 32);
	assert(sizeof(struct nav_navaid_route) == 12);

	latest_filepath[0] = T('\0');

	while (*path) {
		struct directory d;
		const tchar_t *fn = dir_first(*path, &d);

		if (fn != NULL) {
			do {
				static const tchar_t FMT[] = T("%s%c%s");
				struct nav_effective eff;

				tchar_t filepath[1025];
				enum navdb_status result;
				size_t fnlen = tstrlen(fn);

				if (fnlen < 4
				|| (fn[fnlen-1] != 'd' && fn[fnlen-1] != 'D')
				|| (fn[fnlen-2] != 'e' && fn[fnlen-2] != 'E')
				|| fn[fnlen-4] != '.')
					continue;

				tsprintf(filepath, FMT, *path, DIR_SEP, fn);

				result = open_db(filepath, now, &eff);

				if (result == navdb_ok) {
					dir_close(&d);
					err_clear();
					return 1;
				}

				if (result == navdb_err_obsolete) {
					av_usecs s;

					s = av_datetime_to_usecs(eff.start);

					if (s > latest) {
						latest = s;
						tstrncpy(latest_filepath,
							 filepath,
							 ELEMENTS(latest_filepath));
					}
				}
			} while ((fn = dir_next(&d)) != NULL);

			dir_close(&d);
		}

		path++;
	}

	if (latest_filepath[0] != T('\0')) {
		enum navdb_status result;

		result = open_db(latest_filepath, now, NULL);

		if (result == navdb_err_obsolete || result == navdb_ok) {
			err_clear();
			return 1;
		}
	}

	err_push(NAVDB_INIT, NULL, NAVDB_NOT_FOUND);

	return 0;
}

static enum navdb_status
open_db(const tchar_t *fn, av_usecs now,
	struct nav_effective *p_effective_dates) {

	enum navdb_status result;

	if (!map_file(fn, &s_mf)) {
		result = navdb_err_system;
	}
	else {
		result = load_db(s_mf.pv, s_mf.len, now, p_effective_dates);

		if (result != navdb_ok) {
			err_push(NULL, fn, navdb_strerror(result));
			unmap_file(&s_mf);
		}
	}

	return result;
}

static enum navdb_status
load_db(const void *pv, size_t len, av_usecs now,
	struct nav_effective *p_effective_dates) {
	const char *p = (const char *) pv;
	const char *e = p + len;
	enum db i;

	memset((void *) s_p_dbs, 0, sizeof(s_p_dbs));

	while (p < e) {
		static const char TAG[] = "nav-";

		const struct db_header *ph;
		const struct db_trailer *pt;
		size_t bytes;
		unsigned j;

		ph = (const struct db_header *) p;

		if (memcmp(ph->tag, TAG, 4) != 0)
			return navdb_err_not_nav;

		if (ph->endian_check != 0x01234567)
			return navdb_err_endian;

		p += sizeof(struct db_header);
		bytes = ph->num_records * ph->record_length;

		pt = (const struct db_trailer *) (p + bytes);

		if ((const char *) pt < p || (const char *) pt >= e)
			return navdb_err_corrupt;

		if (pt->crc32 != crc32(0, (unsigned char *) p, bytes))
			return navdb_err_crc;

		if (memcmp(pt->tag, TAG, 4) != 0
		    || pt->num_records != ph->num_records
		    || pt->record_length != ph->record_length
		    || memcmp(pt->name, ph->name, sizeof(pt->name)) != 0)
			return navdb_err_corrupt;

		/* look for table name */
		for (j = 0; j < ELEMENTS(TABLES); j++)
			if (strcmp(TABLES[j].name, ph->name) == 0)
				break;

		if (j == ELEMENTS(TABLES)) {
			/* not found, skip */
			p = (const char *) (pt + 1);
			continue;
		}

		if (TABLES[j].record_length != ph->record_length)
			return navdb_err_version;

		i = TABLES[j].db;

		s_p_dbs[i] = p;
		s_e_dbs[i] = (const char *) pt;

		if (i == db_effective) {
			/*
			 * check effective dates against that
			 * provided by user.
			 */
			const struct nav_effective *peff;

			peff = (const struct nav_effective *) p;

			if (p_effective_dates != NULL) {
				*p_effective_dates = *peff;

				if (now < av_datetime_to_usecs(peff->start))
					return navdb_err_future;

				if (av_datetime_to_usecs(peff->end) < now)
					return navdb_err_obsolete;
			}
		}

		p = (const char *) (pt + 1);
	}

	if (s_p_dbs[db_effective] == NULL)
		return navdb_err_missing;

	return navdb_ok;
}

static const tchar_t *
navdb_strerror(enum navdb_status c) {
	switch (c) {
	case navdb_err_directory:
		return ERR_DIRECTORY;

	case navdb_err_system:
		return ERR_SYSTEM;

	case navdb_no_files:
		return NO_FILES;

	case navdb_err_not_nav:
		return ERR_NOT_NAV;

	case navdb_err_endian:
		return ERR_ENDIAN;

	case navdb_err_crc:
		return ERR_CRC;

	case navdb_err_corrupt:
		return ERR_CORRUPT;

	case navdb_err_obsolete:
		return ERR_OBSOLETE;

	case navdb_err_future:
		return ERR_FUTURE;

	case navdb_err_version:
		return ERR_VERSION;

	case navdb_err_missing:
		return ERR_MISSING;

	default:
		return ERR_UNKNOWN;
	}
}

void
navdb_get_effective(struct av_datetime *dt_start,
		    struct av_datetime *dt_end) {
	const struct nav_effective *p;

	p = (const struct nav_effective *) s_p_dbs[db_effective];

	*dt_start = p->start;
	*dt_end = p->end;
}

static const char WP_FMT[] = "WP-%d";

void
navdb_add_waypoint(struct nav_position pos) {
	struct nav_fix *p = s_waypoints + s_iwaypoint;

	s_iwaypoint = (s_iwaypoint + 1) % NUM_WAYPOINTS;
	s_nwaypoints++;

	sprintf(p->ident, "WP-%d", s_nwaypoints);

	p->id = nav_t_fix << 24;
	p->ilat = (int) (pos.latitude / PI * 2147483647.0);
	p->ilon = (int) (pos.longitude / PI * 2147483647.0);
}

void
navdb_enum_ident(const char *ident,
		 void (*pfn)(void *pv, const struct nav_item *p),
		 void *pv) {
	const struct nav_by_ident *p, *e;
	int hi, lo;

	p = (const struct nav_by_ident *) s_p_dbs[db_by_ident];
	e = (const struct nav_by_ident *) s_e_dbs[db_by_ident];

	if (p == e)
		return;

	lo = 0;
	hi = e - p - 1;

	while (lo <= hi) {
		int mid = lo + (hi - lo) / 2;
		int c = strncmp(ident, p[mid].ident, sizeof(p[mid].ident));

		if (c <= 0)
			hi = mid - 1;
		else
			lo = mid + 1;
	}

	while (strncmp(ident, p[lo].ident, sizeof(p[lo].ident)) == 0) {
		const struct nav_item *pni = get_pointer(p[lo].id);

		(*pfn)(pv, pni);

		lo++;
	}

	hi = s_nwaypoints < NUM_WAYPOINTS ? s_nwaypoints : NUM_WAYPOINTS;
	for (lo = 0; lo < hi; lo++)
		if (strncmp(ident, s_waypoints[lo].ident, sizeof(s_waypoints[lo].ident)) == 0)
			(*pfn)(pv, (const struct nav_item *) (s_waypoints + lo));
}

struct nearest_context {
	struct nav_position position;
	const struct nav_item *pnavitem;
	struct nav_polar coords;
};

const struct nav_item *
navdb_find_ident(const char *ident, struct nav_position pos) {
	struct nearest_context c;

	c.position = pos;
	c.pnavitem = NULL;
	c.coords.distance = 1.0e20;
	navdb_enum_ident(ident, nearest, &c);
	return c.pnavitem;
}

void
navdb_enum_prefix(const char *prefix,
		  void (*pfn)(void *pv, const struct nav_item *p),
		  void *pv) {
	const struct nav_by_ident *p, *e;
	int hi, lo;
	size_t l;

	p = (const struct nav_by_ident *) s_p_dbs[db_by_ident];
	e = (const struct nav_by_ident *) s_e_dbs[db_by_ident];

	if (p == e)
		return;

	lo = 0;
	hi = e - p - 1;

	while (lo <= hi) {
		int mid = lo + (hi - lo) / 2;
		int c = strncmp(prefix, p[mid].ident, sizeof(p[mid].ident));

		if (c <= 0)
			hi = mid - 1;
		else
			lo = mid + 1;
	}

	l = strlen(prefix);
	while (strncmp(prefix, p[lo].ident, l) == 0) {
		const struct nav_item *pni = get_pointer(p[lo].id);

		(*pfn)(pv, pni);

		lo++;
	}

	hi = s_nwaypoints < NUM_WAYPOINTS ? s_nwaypoints : NUM_WAYPOINTS;
	for (lo = 0; lo < hi; lo++)
		if (strncmp(prefix, s_waypoints[lo].ident, l) == 0)
			(*pfn)(pv, (const struct nav_item *) (s_waypoints + lo));
}

const struct nav_item *
navdb_find_nearest_by_prefix(const char *prefix, struct nav_position pos) {
	struct nearest_context c;

	c.position = pos;
	c.pnavitem = NULL;
	c.coords.distance = 1.0e20;
	navdb_enum_prefix(prefix, nearest, &c);
	return c.pnavitem;
}

static void
nearest(void *pv, const struct nav_item *pnavitem) {
	struct nearest_context *pc = (struct nearest_context *) pv;
	struct nav_polar c = great_circle(pc->position,
					  navdb_position(pnavitem));

	if (c.distance < pc->coords.distance) {
		pc->pnavitem = pnavitem;
		pc->coords = c;
	}
}

static const struct nav_item *
get_pointer(unsigned id) {
	size_t cb;
	const char *pb;

	switch (id >> 24) {
	case nav_t_airport:
		cb = sizeof(struct nav_airport);
		pb = s_p_dbs[db_airports];
		break;

	case nav_t_runway:
		cb = sizeof(struct nav_runway);
		pb = s_p_dbs[db_runways];
		break;

	case nav_t_navaid:
		cb = sizeof(struct nav_navaid);
		pb = s_p_dbs[db_navaids];
		break;

	case nav_t_fix:
		cb = sizeof(struct nav_fix);
		pb = s_p_dbs[db_fixes];
		break;

	case nav_t_obstruction:
		cb = sizeof(struct nav_obstruction);
		pb = s_p_dbs[db_obstructions];
		break;

	case nav_t_route:
		cb = sizeof(struct nav_route);
		pb = s_p_dbs[db_routes];
		break;

	default:
		return NULL;
	}

	return (const struct nav_item *)(pb + (id & 0x00FFFFFF) * cb);
}

void
navdb_enum_nearest(struct nav_position pos, double radius,
		   void (*pfn)(void *pv, const struct nav_item *p,
			       struct nav_polar *polar),
		   void *pv) {

	const struct nav_by_latitude *p, *e;
	double r = RADIANS(radius / 60.0);

	double minlat = pos.latitude - r;
	double maxlat = pos.latitude + r;

	double minlon, maxlon;

	int iminlat, imaxlat, iminlon, imaxlon;
	int lo, hi;

	if (minlat <= -PI / 2 || PI/2 <= maxlat) {
		minlon = -PI;
		maxlon = PI;
	}
	else {
		double j = r / cos(pos.latitude < 0.0 ? minlat : maxlat);

		minlon = pos.longitude - j;
		maxlon = pos.longitude + j;
	}

	p = (const struct nav_by_latitude *) s_p_dbs[db_by_latitude];
	e = (const struct nav_by_latitude *) s_e_dbs[db_by_latitude];

	if (p == e)
		return;

	iminlat = (int) (minlat / PI * 2147483647.0);
	imaxlat = (int) (maxlat / PI * 2147483647.0);
	iminlon = (int) (minlon / PI * 2147483647.0);
	imaxlon = (int) (maxlon / PI * 2147483647.0);

	/* binary search for first item in list */
	lo = 0;
	hi = e - p - 1;
	while (lo <= hi) {
		unsigned mid = lo + (hi - lo) / 2;

		if (iminlat < p[mid].ilat)
			hi = mid - 1;
		else
			lo = mid + 1;
	}

	/* p[lo] is the southern-most candidate */
	p += lo;
	while (p < e && p->ilat <= imaxlat) {
		const struct nav_item *pi = get_pointer(p->id);

		if (iminlon <= pi->ilon && pi->ilon <= imaxlon) {
			struct nav_position posn = navdb_position(pi);
			struct nav_polar polar = great_circle(pos, posn);

			if (polar.distance < radius)
				(*pfn)(pv, pi, &polar);
		}

		p++;
	}

	hi = s_nwaypoints < NUM_WAYPOINTS ? s_nwaypoints : NUM_WAYPOINTS;
	for (lo = 0; lo < hi; lo++) {
		const struct nav_item *pi;

		pi = (const struct nav_item *) (s_waypoints + lo);

		if (iminlat <= pi->ilat && pi->ilat <= imaxlat &&
		    iminlon <= pi->ilon && pi->ilon <= imaxlon) {
			struct nav_position posn = navdb_position(pi);
			struct nav_polar polar = great_circle(pos, posn);
			if (polar.distance < radius)
				(*pfn)(pv, pi, &polar);
		}
	}
}

const struct nav_item *
navdb_other_runway_side(const struct nav_item *p) {
	unsigned id = p->id ^ 1;

	return get_pointer(id);
}


/**
 * Retrieves text description of nav item.
 *
 * If _pnav_ is NULL, sets buf to an empty string and returns 0.
 * If _pnav_ points to an airport, fix, or navaid, retrieves the
 * ident and copies it into buf, returning the string length.
 * If _pnav_ is an obstruction, formats the _msl_ into buf.
 */
unsigned
navdb_get_ident_a(const struct nav_item *pnav, char *dst, unsigned dstlen) {
	const char *src;
	unsigned srclen, i;
	char buf[8];

	get_ident(pnav, buf, &src, &srclen);

	if (srclen >= dstlen)
		srclen = dstlen - 1;

	for (i = 0; i < srclen; i++)
		if ((dst[i] = src[i]) == '\0')
			break;

	dst[i] = '\0';

	return i;
}

#ifdef UNICODE
unsigned
navdb_get_ident_u(const struct nav_item *pnav, wchar_t *dst, unsigned dstlen) {
	const char *src;
	unsigned srclen, i;
	char buf[16];

	get_ident(pnav, buf, &src, &srclen);

	if (srclen >= dstlen)
		srclen = dstlen - 1;

	for (i = 0; i < srclen; i++)
		if ((dst[i] = src[i]) == '\0')
			break;

	dst[i] = '\0';

	return i;
}
#endif

static void
get_ident(const struct nav_item *pnav, char buf[16], const char **psrc,
	  unsigned *plen) {
	static const char INT_FMT[] = "%d";

	if (pnav == NULL)
		*plen = 0;
	else
		switch (NAV_TYPE(pnav)) {
		case nav_t_airport:
			*psrc = ((const struct nav_airport *) pnav)->ident;
			*plen = sizeof ((const struct nav_airport *) pnav)->ident;
			break;

		case nav_t_runway:
			*psrc = ((const struct nav_runway *) pnav)->ident;
			*plen = sizeof ((const struct nav_runway *) pnav)->ident;
			break;

		case nav_t_navaid:
			*psrc = ((const struct nav_navaid *) pnav)->ident;
			*plen = sizeof ((const struct nav_navaid *) pnav)->ident;
			break;

		case nav_t_fix:
			*psrc = ((const struct nav_fix *) pnav)->ident;
			*plen = sizeof ((const struct nav_fix *) pnav)->ident;
			break;

		case nav_t_obstruction:
			*plen = sprintf(buf, INT_FMT,
					((const struct nav_obstruction *) pnav)->msl);
			*psrc = buf;
			break;

		case nav_t_route:
			*psrc = ((const struct nav_route *) pnav)->name;
			*plen = sizeof ((const struct nav_route *) pnav)->name;
			break;

		default:
			*plen = 0;
		}
}

struct nav_position
navdb_position(const struct nav_item *pnav) {
	struct nav_position pos;

	pos.latitude = IRADIANS(pnav->ilat);
	pos.longitude = IRADIANS(pnav->ilon);

	return pos;
}

unsigned
navdb_fix_routes(const struct nav_item *pfix,
		 const struct nav_item *routes[],
		 const struct nav_item *fixes[],
		 unsigned nroutes) {
	const struct nav_navaid_route *p, *e;
	int lo, hi;

	unsigned i = 0, id = pfix->id;

	p = (const struct nav_navaid_route *) s_p_dbs[db_navaid_routes];
	e = (const struct nav_navaid_route *) s_e_dbs[db_navaid_routes];

	if (p == e)
		return 0;

	/* binary search for first item in list */
	lo = 0;
	hi = e - p - 1;
	while (lo <= hi) {
		unsigned mid = lo + (hi - lo) / 2;

		if (id <= p[mid].id)
			hi = mid - 1;
		else
			lo = mid + 1;
	}

	p += lo;
	while (p < e && p->id == id && i < nroutes) {
		routes[i] = get_pointer(p->route_id);
		fixes[i] = get_pointer(p->to_id);
		i++;
		p++;
	}

	while (p < e && p->id == id) {
		i++;
		p++;
	}

	return i;
}

/*
 * Overrides for Emacs so that we follow Linux's tabbing style.
 * Emacs will notice this stuff at the end of the file and automatically
 * adjust the settings for this buffer only.  This must remain at the
 * end of the file.
 * ----------------------------------------------------------------------
 * Local Variables:
 * c-file-style: "linux"
 * End:
 */
