/**
 * http://www.wayforward.net/avGPS/libav/av_err.c
 * Copyright(c) 2005, 2008 Wayforward LLC
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307
 * USA
 */
#ifdef _WIN32
# define WIN32_LEAN_AND_MEAN
# include <windows.h>

#else
# include <stdio.h>
# include <errno.h>
# include <unistd.h>

#endif

#include <stdlib.h>
#include <string.h>

#include "av_err.h"

static void
push_str(const tchar_t *str);

static void
push_char(tchar_t ch);

#define CWDBUFLEN (256)

static tchar_t error_text[80 * 12 + CWDBUFLEN];
static tchar_t *error_end, *error_start;

void
err_init(void) {
	error_end = error_text + ELEMENTS(error_text) - CWDBUFLEN;
	error_start = error_end;
}

void
err_push(const tchar_t *function, const tchar_t *object,
	 const tchar_t *reason) {
	/* add things to the -front- of error text */
	push_char(T('\n'));
	push_char(T('\n'));

	if (object != NULL || function != NULL) {
		push_char(T(')'));

		if (object != NULL) {
			push_char(T('"'));
			push_str(object);
			push_char(T('"'));

			if (function != NULL)
				push_char(T(' '));
		}

		if (function != NULL)
			push_str(function);

		push_char(T('('));
	}

	if (tstrlen(reason) > 0
	    && reason[tstrlen(reason) - 1] != T('\n'))
		push_char(T('\n'));

	push_str(reason);
}

static void
push_str(const tchar_t *str) {
	size_t l = tstrlen(str);

	if (error_start - l >= error_text) {
		error_start -= l;
		memcpy(error_start, str, l * sizeof(tchar_t));
	}
}

static void
push_char(tchar_t ch) {
	if (error_start > error_text)
		*--error_start = ch;
}

void
err_push_source(const tchar_t *filename, unsigned long lineno,
		const tchar_t *reason) {
	if (tstrlen(reason) > 0
	    && reason[tstrlen(reason) - 1] != T('\n'))
		push_char(T('\n'));

	push_str(reason);
	push_char(T(' '));
	push_char(T(':'));

	if (lineno > 0) {
		static const tchar_t FMT[] = T("(%lu)");

		tchar_t buf[16];
		tsprintf(buf, FMT, lineno);
		push_str(buf);
	}

	push_str(filename);
}

#ifdef _WIN32
void
err_push_last_error(const tchar_t *function, const tchar_t *object) {
	err_push_win32_error(function, object, GetLastError());
}

void
err_push_win32_error(const tchar_t *function, const tchar_t *object,
		     unsigned long error) {
	tchar_t buf[1025];
	DWORD dwResult;

	dwResult = FormatMessage(FORMAT_MESSAGE_FROM_SYSTEM, NULL,
			  error, 0,
			  buf, ELEMENTS(buf), NULL);
	
	if (dwResult == 0) {
		static const tchar_t FMT[] = T("Error %d (0x%08x)");
		tsprintf(buf, FMT, error, error);
	}

	err_push(function, object, buf);
}

#else /* !_WIN32 */

void
err_push_errno(const tchar_t *function, const tchar_t *object) {
	err_push_c_error(function, object, errno);
}

void
err_push_c_error(const tchar_t *function, const tchar_t *object, int err) {
	err_push(function, object, strerror(err));
}

#endif /* _WIN32 */

void
err_exit(const tchar_t *desc) {
#ifndef _WIN32_WCE
	static const tchar_t CURRENT_DIRECTORY[] = T("Current directory: ");

	tstrcpy(error_end, CURRENT_DIRECTORY);

# ifdef _WIN32
	GetCurrentDirectory(CWDBUFLEN - tstrlen(error_end),
			    error_end + tstrlen(error_end));
# else
	getcwd(error_end + tstrlen(error_end),
	       CWDBUFLEN - tstrlen(error_end));
# endif
#endif

	push_char(T('\n'));
	push_char(T('\n'));
	push_str(desc);

#ifdef _WIN32
	MessageBox(NULL, error_start, NULL, MB_OK | MB_ICONERROR);
	ExitProcess(1);

#else
	fputs(error_start, stderr);
	putc('\n', stderr);
	exit(EXIT_FAILURE);
#endif

	/* NOTREACHED */
}

const tchar_t *
err_text(void) {
	return error_start;
}

void
err_clear(void) {
	error_start = error_end;
	*error_start = T('\0');
}

/*
 * Overrides for Emacs so that we follow Linux's tabbing style.
 * Emacs will notice this stuff at the end of the file and automatically
 * adjust the settings for this buffer only.  This must remain at the
 * end of the file.
 * ----------------------------------------------------------------------
 * Local Variables:
 * c-file-style: "linux"
 * End:
 */
