/*
 * http://www.wayforward.net/avGPS/libav/av_gps.h
 * Copyright(c) 2005, 2008 Wayforward LLC
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307
 * USA
 */
#ifndef AV_GPS_H
#define AV_GPS_H

#include <stdio.h>

#include "av_comm.h"
#include "av_time.h"

#ifdef __cplusplus
extern "C" {
#endif

	struct gps_satellite {
		int satellite_id;
		int elevation_degrees;
		int azimuth_degrees;
		int snr_dbhz;
	};

	struct gps_device {
		struct comm_device device;

		/* saved satellite quality/signal strength */
		int quality;
		struct gps_satellite sats[24];
		int nsats;

		/* saved course, for when RMC heading is blank */
		double true_course;
	};

	enum gps_type {
		gps_t_fix,
		gps_t_course,
		gps_t_minimum,
		gps_t_active,
		gps_t_satellites,
		gps_t_quality
	};

	enum gps_fix_quality {
		gps_quality_invalid = 0,/* no signal			*/
		gps_quality_sps  = 1,	/* civilian GPS			*/
		gps_quality_dgps = 2,	/* differential GPS		*/
		gps_quality_pps = 3,	/* military GPS			*/
		gps_quality_rtk = 4,	/* real-time kinematic		*/
		gps_quality_frtk = 5,	/* float real-time kinematic	*/
		gps_quality_est = 6,	/* estimated or dead-reckoning	*/
		gps_quality_manual = 7,	/* manual input mode		*/
		gps_quality_sim = 8,	/* simulated			*/
	};

	struct gps_fix {
		enum gps_type	     type;
		struct av_time	     time;
		enum gps_fix_quality quality;
		double		     latitude;
		double		     longitude;
		double		     msl_meters;
	};

	struct gps_course {
		enum gps_type	type;
		double		true_course;
		double		knots;
		int             true_course_valid;
		int             knots_valid;
	};

	struct gps_minimum {
		enum gps_type	type;
		struct av_datetime datetime;
		int             valid;
		double		latitude;
		double		longitude;
		double		knots;
		double		true_course;
		int		true_course_valid;
	};

	struct gps_active {
		enum gps_type	type;
		const char *	op_mode;
		const char *	fix_mode;
		int		prns[12];
		unsigned	nprns;
		double		pdop;
		double		hdop;
		double		vdop;
	};

	struct gps_satellites {
		enum gps_type	type;
		const struct gps_satellite *satellites;
		unsigned	nsatellites;
	};

	struct gps_quality {
		enum gps_type	     type;
		enum gps_fix_quality quality;
	};

	union gps_update {
		enum gps_type		type;
		struct gps_fix		fix;
		struct gps_course	course;
		struct gps_minimum	minimum;
		struct gps_active	active;
		struct gps_satellites	satellites;
		struct gps_quality	quality;
	};

	struct gps_listener {
		void (*start)  (void *pv);

		void (*fix)    (void *pv, const struct gps_fix *pfix);
		void (*course) (void *pv, const struct gps_course *pcourse);
		void (*minimum)(void *pv, const struct gps_minimum *pminimum);
		void (*active) (void *pv, const struct gps_active *pactive);

		void (*satellites)(void *pv,
				   const struct gps_satellites *psatellite);
		void (*quality)(void *pv, const struct gps_quality *pquality);

		void (*end)    (void *pv);

		/* hook for any unparsed, but valid, GPS data */
		void (*unhandled)(void *pv, char **tokens, unsigned n);

		/* hook for application to log the raw input, \n and all */
		void (*log)(void *pv, const char *line, size_t len);
	};

	/* returns 0 on error, 1 on success */
	int
	gps_open(struct gps_device *pd, const tchar_t *devicename);

#define gps_is_live(pd) (comm_is_live(&(pd)->device))
#define gps_is_running(pd) (comm_is_running(&(pd)->device))
#define gps_stop(pd) (comm_stop(&(pd)->device))
#define gps_start(pd) (comm_start(&(pd)->device))

	int
	gps_loop(struct gps_device *pd, const struct gps_listener *p,
		 void *pv);

	void
	gps_close(struct gps_device *pd);

#ifdef __cplusplus
}
#endif

#endif /* AV_GPS_H */

/*
 * Overrides for Emacs so that we follow Linux's tabbing style.
 * Emacs will notice this stuff at the end of the file and automatically
 * adjust the settings for this buffer only.  This must remain at the
 * end of the file.
 * ----------------------------------------------------------------------
 * Local Variables:
 * c-file-style: "linux"
 * End:
 */
