/*
 * http://www.wayforward.net/avGPS/libav/av_lomap.c
 * Copyright(c) 2005, 2008 Wayforward LLC
 *
 * Loads terrain tiles, 1 pixel per 300 feet, or 300 per 1/4 degree.
 * The terrain tiles are in '.dat' format, one pixel every two bytes,
 * suitable for mapping into memory.
 * --------------------------------------------------------------------
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307
 * USA
 */
#include <assert.h>
#include <math.h>
#include <stdlib.h>
#include <string.h>
#include <stdio.h>

#define LORESMAP
#include "av_map.h"

#define INDEX_NORTH (16)
#define INDEX_SOUTH (-16)
#define INDEX_EAST (2)
#define INDEX_WEST (-2)

#include "av_mem.h"

int
index_north(int index) {
	return index + INDEX_NORTH;
}

int
index_south(int index) {
	return index + INDEX_SOUTH;
}

#define TILE_EAST ((1 << 14) - 6)
#define TILE_WEST (-TILE_EAST)

int
index_east(int index) {
	return index + ((index & 7) == 6 ? TILE_EAST : INDEX_EAST);
}

int
index_west(int index) {
	return index + ((index & 6) == 0 ? TILE_WEST : INDEX_WEST);
}

struct map_coords
unpack_index(int index) {
	struct map_coords result;

	result.ilat = (index >>  6 & 255) -  90;
	result.ilon = (index >> 14 & 511) - 180;
	result.x = index & 6;
	result.y = index >> 3 & 6;

	return result;
}

int
pack_index(int ilat, int ilon, int x, int y) {
	return ((ilon + 180) << 14) + ((ilat + 90) << 6) + (y << 3) + x;
}

struct map_index
latlon2index(double lat, double lon) {
	struct map_index result;

	double lat4 = lat * 4.0;
	double lon4 = lon * 4.0;

	int ilat4 = (int) floor(lat4);
	int ilon4 = (int) floor(lon4);

	result.index = pack_index(ilat4 >> 2, ilon4 >> 2,
				  (ilon4 & 0x03) << 1,
				  (ilat4 & 0x03) << 1);
	result.x = lon4 - ilon4;
	result.y = lat4 - ilat4;

	return result;
}

void
index2latlon(int index, double *plat, double *plon) {
	struct map_coords c = unpack_index(index);

	*plon = c.ilon + c.x / 8.0;
	*plat = c.ilat + c.y / 8.0;
}

static const tchar_t FMT_A[] = T("%c%d0");
static const tchar_t FMT_B[] = T("%c%d");
static const tchar_t FMT[] = T("%s/%ce-300x4/%s%s/%s%s/%d");

static tchar_t
get_endian(void) {
	union {
		unsigned char b[2];
		unsigned short s;
	} u;

	u.s = 0x0123;
	if (u.b[0] == 0x01) {
		assert(u.b[1] == 0x23);

		return T('b');
	}
	else {
		assert(u.b[1] == 0x01);
		assert(u.b[0] == 0x23);

		return T('l');
	}
}

void
index_filename(const tchar_t *dir, int index, tchar_t *filename, size_t fnlen) {
	struct map_coords c = unpack_index(index);

	tchar_t dir1a[16];
	tchar_t dir1b[16];

	tchar_t dir2a[16];
	tchar_t dir2b[16];

	if (c.ilon < 0) {
		tsprintf(dir1a, FMT_A, T('w'), (9 - c.ilon) / 10);
		tsprintf(dir2a, FMT_B, T('w'), -c.ilon);
	}
	else {
		tsprintf(dir1a, FMT_A, T('e'), c.ilon / 10);
		tsprintf(dir2a, FMT_B, T('e'), c.ilon);
	}

	if (c.ilat < 0) {
		tsprintf(dir1b, FMT_A, T('s'), (9 - c.ilat)/ 10);
		tsprintf(dir2b, FMT_B, T('s'), -c.ilat);
	}
	else {
		tsprintf(dir1b, FMT_A, T('n'), c.ilat / 10);
		tsprintf(dir2b, FMT_B, T('n'), c.ilat);
	}

	tsprintf(filename, FMT, dir, get_endian(), dir1a, dir1b, dir2a, dir2b,
		 index);
}

static const tchar_t SUFFIX[] = T(".dat");
static const tchar_t MODE[] = T("r");

#define TILE_LENGTH (300 * 300)

static struct cache {
	struct mapped_file mf;
	const short *tile;
	int index;
	int width, height;
} cache[9];

static int cache_length = 0;

const short *
fetch_tile(const tchar_t *dir, int index, int *pw, int *ph) {
	int i;
	const short *result;
	tchar_t filename[512];
	struct mapped_file mf;

	for (i = 0; i < cache_length; i++)
		if (index == cache[i].index) {
			*pw = cache[i].width;
			*ph = cache[i].height;
			return cache[i].tile;
		}

	index_filename(dir, index, filename, ELEMENTS(filename));
	tstrcat(filename, SUFFIX);

	if (!map_file(filename, &mf))
		return NULL;

	assert(mf.len == 300 * 300 * sizeof(short));

	if (cache_length == ELEMENTS(cache))
		unmap_file(&cache[ELEMENTS(cache) - 1].mf);
	else
		cache_length++;

	for (i = 1; i < cache_length; i++)
		cache[i] = cache[i - 1];

	cache[0].mf = mf;
	result = cache[0].tile = (const short *) cache[0].mf.pv;

	cache[0].index = index;
	cache[0].width = *pw = 300;
	cache[0].height = *ph = 300;

	return result;
}

/*
 * Overrides for Emacs so that we follow Linux's tabbing style.
 * Emacs will notice this stuff at the end of the file and automatically
 * adjust the settings for this buffer only.  This must remain at the
 * end of the file.
 * ----------------------------------------------------------------------
 * Local Variables:
 * c-file-style: "linux"
 * End:
 */
