/*
 * http://www.wayforward.net/avGPS/libav/av_nav.c
 * Copyright(c) 2005, 2008 Wayforward LLC
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307
 * USA
 */
#include <math.h>
#include <stdlib.h>
#include <stdio.h>
#include <string.h>

#include "av_nav.h"

static double
parse_angle(const char *str, char **pp);

static tchar_t *
format(tchar_t *p, tchar_t *pe, const tchar_t **pfmt, double coord);

const tchar_t AVDB_LAT_HTML[] = T("%N%D&deg; %05.2m'");
const tchar_t AVDB_LON_HTML[] = T("%E%D&deg; %05.2m'");

#ifdef UNICODE
const tchar_t AVDB_LAT_FMT[] = T("%N%D\x00b0 %05.2m'");
const tchar_t AVDB_LON_FMT[] = T("%E%D\x00b0 %05.2m'");

#else
const tchar_t AVDB_LAT_FMT[] = T("%N%D\xc2\xb0 %05.2m'");
const tchar_t AVDB_LON_FMT[] = T("%E%D\xc2\xb0 %05.2m'");

const char AVDB_LAT_LATIN1[] = T("%N%D\xb0 %05.2m'");
const char AVDB_LON_LATIN1[] = T("%E%D\xb0 %05.2m'");

#endif

struct nav_cartesian
cartesian(double a, double d) {
	struct nav_cartesian r;

	r.x = sin(a) * d;
	r.y = cos(a) * d;

	return r;
}

struct nav_polar
polar(double x, double y) {
	struct nav_polar r;

	r.angle = atan2(x, y);
	r.distance = sqrt(x * x + y * y);

	return r;
}

static double
p2(double d) {
	return d * d;
}

/**
 * Calculate distance (in nautical miles) and true course (in radians)
 * between two points on the Earth, using a great circle path.
 *
 * Returns (nm, radians)
 */
struct nav_polar
great_circle(struct nav_position np1, struct nav_position np2) {
	double lat1 = np1.latitude, lon1 = np1.longitude;
	double lat2 = np2.latitude, lon2 = np2.longitude;

	double cos_lat1 = cos(lat1), d, tc1;
	struct nav_polar r;

	d = 2.0 * asin(sqrt(p2(sin((lat1 - lat2) / 2.0)) +
			    cos_lat1 * cos(lat2) *
			    p2(sin((lon2 - lon1) / 2.0))));

	if (d == 0.0)
		r.angle = r.distance = 0.0;
	else {
		tc1 = (sin(lat2) - sin(lat1) * cos(d)) / (sin(d) * cos_lat1);

		if (tc1 > 1.0) tc1 = 1.0;
		else if (tc1 < -1.0) tc1 = -1.0;
		tc1 = acos(tc1);

		if (sin(lon1 - lon2) >= 0)
			tc1 = 2 * PI - tc1;

		r.angle = tc1;
		r.distance = 10800.0 * d / PI;
	}

	return r;
}

double
nav_parse_latitude(const char *string) {
	char *p;
	double degrees, sign = 1.0;

	if (*string == 's' || *string == 'S') {
		string++;
		sign = -1.0;
	}
	else if (*string == 'n' || *string == 'N')
		string++;

	degrees = sign * parse_angle(string, &p);

	if (*p == 's' || *p == 'S') {
		p++;
		sign = -1.0;
	}
	else if (*p == 'n' || *p == 'N')
		p++;

	return *p == '\0' ? sign * degrees : INVALID_LATITUDE;
}

double
nav_parse_longitude(const char *string) {
	char *p;
	double degrees, sign = 1.0;

	if (*string == 'w' || *string == 'W') {
		string++;
		sign = -1.0;
	}
	else if (*string == 'e' || *string == 'E')
		string++;

	degrees = sign * parse_angle(string, &p);

	if (*p == 'w' || *p == 'W') {
		p++;
		sign = -1.0;
	}
	else if (*p == 'e' || *p == 'E')
		p++;

	return *p == '\0' ? sign * degrees : INVALID_LONGITUDE;
}

static double
parse_angle(const char *str, char **pp) {
	double degrees = strtod(str, pp);

	if (**pp == ':') {
		double minutes = strtod(*pp + 1, pp);

		if (**pp == ':') {
			double seconds = strtod(*pp + 1, pp);

			minutes += seconds / 60.0;
		}

		degrees += minutes / 60.0;
	}

	return degrees;
}

/**
 * Use a format string to control how a coordinate (latitude
 * or longitude) is displayed.
 *
 * The format string is copied verbatim, except for characters
 * preceded by a percent (%) sign.  The format characters are:
 *    %%    - insert a single percent sign
 *    %D    - integer degrees
 *    %d    - decimal degrees
 *    %M    - integer minutes
 *    %m    - decimal minutes
 *    %S    - integer seconds
 *    %s    - decimal seconds
 *    %n    - 'n' if coord is positive, 's' if negative
 *    %N    - 'N' if coord is positive, 'S' if negative
 *    %e    - 'e' if coord is positive, 'w' if negative
 *    %E    - 'E' if coord is positive, 'W' if negative
 *    %U    - 'E' if coord is positive, 'U' if negative
 *    %+    - '+' if coord is positive, '-' if negative
 *    %-    - blank if coord is positive, '-' if negative
 *
 * The dDmMsS format characters can be preceded by printf-style
 * numbers.
 *
 * Examples:
 * "%02D:%02M:%02S:%U"      :  121:27:54:E  for Lishe, Ninbo, China
 */
size_t
strfcoord(tchar_t *buf, size_t nbuf, const tchar_t *fmt, double coord) {
	tchar_t *p = buf, *pe = buf + nbuf;
	int neg = (coord < 0.0);

	if (neg)
		coord = -coord;

	while (p != pe) {
		tchar_t ch = *fmt++;

		switch (ch) {
		case T('%'):
			switch (*fmt++) {
			case T('%'):
				*p++ = ch;
				break;

			case T('E'):
				*p++ = neg ? T('W') : T('E');
				break;

			case T('e'):
				*p++ = neg ? T('w') : T('e');
				break;

			case T('N'):
				*p++ = neg ? T('S') : T('N');
				break;

			case T('n'):
				*p++ = neg ? T('s') : T('n');
				break;

			case T('U'):
				*p++ = neg ? T('U') : T('E');
				break;

			case T('+'):
				*p++ = neg ? T('-') : T('+');
				break;

			case T('-'):
				if (neg) *p++ = T('-');
				break;

			default:
				p = format(p, pe, &fmt, coord);
			}
			break;

		case T('\0'):
			*p = ch;
			return p - buf;

		default:
			*p++ = ch;
		}
	}

	return nbuf;
}

static double
frac(double d) {
	return d - floor(d);
}

static const tchar_t ARG_CHARS[] = T("0123456789.");

static tchar_t *
format(tchar_t *p, tchar_t *pe, const tchar_t **pfmt, double coord) {
	const tchar_t *fmt = *pfmt - 1;
	tchar_t text[20], format[20];
	unsigned n = 0;

	format[n++] = T('%');

	while (tstrchr(ARG_CHARS, *fmt) != NULL)
		format[n++] = *fmt++;

	switch (*fmt++) {
	case T('D'):
		/* %D    - integer degrees */
		format[n++] = T('g');
		coord = floor(coord);
		break;

	case T('d'):
		/* %d    - decimal degrees */
		format[n++] = T('f');
		break;

	case T('M'):
		/* %M    - integer minutes */
		format[n++] = T('g');
		coord = fmod(floor(coord * 60.0), 60.0);
		break;

	case T('m'):
		/* %m    - decimal minutes */
		format[n++] = T('f');
		coord = frac(coord) * 60.0;
		break;

	case T('S'):
		/* %S    - integer seconds */
		format[n++] = T('g');
		coord = fmod(floor(coord * 3600.0), 60.0);
		break;

	case T('s'):
		/* %s    - decimal seconds */
		format[n++] = T('f');
		coord = frac(coord * 60.0) * 60.0;
		break;

	default:
		return p;
	}

	format[n] = T('\0');
#ifdef UNICODE
	swprintf(text, format, coord);
#else
	sprintf(text, format, coord);
#endif

	{
		tchar_t *pt = text;

		while (p != pe) {
			if ((*p = *pt++) == T('\0'))
				break;

			p++;
		}
	}

	*pfmt = fmt;
	return p;
}


/*
 * Overrides for Emacs so that we follow Linux's tabbing style.
 * Emacs will notice this stuff at the end of the file and automatically
 * adjust the settings for this buffer only.  This must remain at the
 * end of the file.
 * ----------------------------------------------------------------------
 * Local Variables:
 * c-file-style: "linux"
 * End:
 */
