/*
 * http://www.wayforward.net/avGPS/include/av_nav.h
 * Copyright(c) 2005, 2008 Wayforward LLC
 *
 * Navigation on a sphere
 * Utility routines for 2D cartesian to polar conversions.
 * Formatting and parsing latitude/longitudes
 * ---------------------------------------------------------------------
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307
 * USA
 */
#ifndef AV_NAV_H
#define AV_NAV_H

#include <stdlib.h>

#include "av_str.h"

#ifndef PI
# define PI (3.14159265358979323)
#endif

/* convert degrees to radians */
#define RADIANS(d) ((d) / (180.0 / PI))

/* convert radians to degrees */
#define DEGREES(r) ((r) * (180.0 / PI))

#ifdef __cplusplus
extern "C" {
#endif

#define INVALID_LATITUDE (720.0)
#define INVALID_LONGITUDE (720.0)

	struct nav_cartesian {
		double x, y;
	};

	struct nav_polar {
		double angle, distance;
	};

	struct nav_position {
		double latitude, longitude;
	};

	/**
	 * Returns cartesian x, y coordinates given an angle _a_ and distance
	 * _d_.
	 *
	 * Angle _a_ is in radians.
	 *
	 * x = sin(a) * d
	 * y = cos(a) * d
	 */
	struct nav_cartesian
	cartesian(double a, double d);

	/**
	 * Returns polar angle, distance given cartesian _x_, _y_
	 * coordinates.
	 */
	struct nav_polar
	polar(double x, double y);

	/*
	 * Given two latitude/longitude coordinates, calculate the great
	 * circle heading from _p1_ to _p2_.  Heading returned is in
	 * radians from north.  Distance returned is in nautical miles.
	 */
	struct nav_polar
	great_circle(struct nav_position p1, struct nav_position p2);

	/**
	 * Parse strings of the form
	 *    ddd.dd      - decimal degrees
	 *    ddd:mm.mmm  - integer degrees and decimal minutes
	 *    ddd:mm:ss.s - integer degrees and minutes, decimal seconds
	 *
	 * If the string is prefixed or suffixed by 's' or 'S', the
	 * result is negative.  Similarly, 'n' or 'N' prefix or suffix
	 * results in a positive value.
	 *
	 * Returns INVALID_LATITUDE on parse error.
	 */
	double
	nav_parse_latitude(const char *string);

	/**
	 * Parse strings of the form
	 *    ddd.dd      - decimal degrees
	 *    ddd:mm.mmm  - integer degrees and decimal minutes
	 *    ddd:mm:ss.s - integer degrees and minutes, decimal seconds
	 *
	 * If the string is prefixed or suffixed by 'w' or 'W', the
	 * result is negative.  Similarly, 'e' or 'E' prefix or suffix
	 * results in a positive value.
	 *
	 * Returns INVALID_LONGITUDE on parse error.
	 */
	double
	nav_parse_longitude(const char *string);

	/**
	 * Use a format string to control how a coordinate (latitude
	 * or longitude) is displayed.
	 *
	 * The format string is copied verbatim, except for characters
	 * preceded by a percent (%) sign.  The format characters are:
	 *    %%    - insert a single percent sign
	 *    %D    - integer degrees
	 *    %d    - decimal degrees
	 *    %M    - integer minutes
	 *    %m    - decimal minutes
	 *    %S    - integer seconds
	 *    %s    - decimal seconds
	 *    %n    - 'n' if coord is positive, 's' if negative
	 *    %N    - 'N' if coord is positive, 'S' if negative
	 *    %e    - 'e' if coord is positive, 'w' if negative
	 *    %E    - 'E' if coord is positive, 'W' if negative
	 *    %U    - 'E' if coord is positive, 'U' if negative
	 *    %+    - '+' if coord is positive, '-' if negative
	 *    %-    - blank if coord is positive, '-' if negative
	 *
	 * The dDmMsS format characters can be preceded by printf-style
	 * numbers.
	 *
	 * Examples:
	 * "%02D:%02M:%02S:%U"      :  121:27:54:E  for Lishe, Ninbo, China
	 *
	 * Use AVDB_LAT_FMT for the standard AF/D latitude format,
	 * Use AVDB_LON_FMT for the standard AF/D longitude format.
	 */
	size_t
	strfcoord(tchar_t *buf, size_t nbuf, const tchar_t *fmt, double coord);

	extern const tchar_t AVDB_LAT_HTML[], AVDB_LON_HTML[];
	extern const tchar_t AVDB_LAT_FMT[], AVDB_LON_FMT[];
#ifndef UNICODE
	extern const char AVDB_LAT_LATIN1[], AVDB_LON_LATIN1[];
#endif

#ifdef __cplusplus
}
#endif

#endif /* AV_NAV_H */

/*
 * Overrides for Emacs so that we follow Linux's tabbing style.
 * Emacs will notice this stuff at the end of the file and automatically
 * adjust the settings for this buffer only.  This must remain at the
 * end of the file.
 * ----------------------------------------------------------------------
 * Local Variables:
 * c-file-style: "linux"
 * End:
 */
