#include <assert.h>
#include <math.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include <sgb/gb_graph.h>
#include <sgb/gb_save.h>

#ifndef PI
# define PI (3.14159265358979323)
#endif

/* convert degrees to radians */
#define RADIANS(d) ((d) * (PI / 180.0))

/* convert radians to degrees */
#define DEGREES(r) ((r) * (180.0 / PI))

/*
 * Compile fixes.csv, routes.csv, and legs.csv into a Stanford Graphbase
 * (Donald Knuth's literate programming with graphs).
 *
 * An SGB Graph hold the entire US navigation route structure
 *    uu: effective start date
 *    vv: effective end date
 *    ww: software version
 *    xx: NOT FOR NAVIGATION
 *    yy:
 *    zz:
 *
 * An SGB Vertex is an avdata Fix.
 *    name:   faa identifier of fix
 *    u: internal sgb hashing
 *    v: internal sgb hashing
 *    w: latitude of fix in arcseconds
 *    x: longitude of fix in arcseconds
 *    y: internal sgb Dijkstra
 *    z: internal sgb Dijkstra
 *
 * An SGB Arc is an avdata leg.
 *    len:   length of arc (in arcseconds, 1/60th of a nautical mile)
 *    a:     name of route
 *    b:
 *
 */
int parse_csv(char *line, char **columns, int ncolumns);

double
great_circle(double lat1, double lon1, double lat2, double lon2);

double p2(double d);

#define MAX_COLUMNS (64)

int main(int argc, char **argv) {
    FILE *fp;
    Graph *g;
    Vertex *v;
    long lineno = 0;
    char *columns[MAX_COLUMNS];
    char buf[512];

    if (argc != 3) {
        fprintf(stderr, "Usage: av_sgb {csv dir} {routes.gb filename}\n");
        exit(EXIT_FAILURE);
    }

    sprintf(buf, "%s/fixes.csv", argv[1]);

    /* count number of vertices (number of fixes in fixes.csv) */
    fp = fopen(buf, "r");
    if (fp == NULL) {
        perror(buf);
        exit(EXIT_FAILURE);
    }

    while (fgets(buf, sizeof(buf), fp))
        lineno++;

    fclose(fp);

    g = gb_new_graph(lineno - 2);
    if (g == NULL) {
        fprintf(stderr, "gb_new_graph: couldn't create graph with %ld vertices\n",
                lineno - 2);
        exit(EXIT_FAILURE);
    }

    strcpy(g->util_types, "VVIIZZSZSSSSZZ");

    sprintf(buf, "%s/fixes.csv", argv[1]);
    fp = fopen(buf, "r");
    if (fp == NULL) {
        perror(buf);
        exit(EXIT_FAILURE);
    }

    fgets(buf, sizeof(buf), fp);
    parse_csv(buf, columns, MAX_COLUMNS);
    assert(strncmp(columns[0], "start=", 6) == 0);
    assert(strncmp(columns[1], "end=", 4) == 0);
    assert(strncmp(columns[2], "version=", 8) == 0);
    assert(strncmp(columns[3], "note=", 5) == 0);
    g->uu.S = gb_save_string(columns[0]+6);
    g->vv.S = gb_save_string(columns[1]+4);
    g->ww.S = gb_save_string(columns[2]+8);
    g->xx.S = gb_save_string(columns[3]+5);

    fgets(buf, sizeof(buf), fp);

    v = g->vertices;
    while (fgets(buf, sizeof(buf), fp)) {
        int n = parse_csv(buf, columns, MAX_COLUMNS);

        v->name = gb_save_string(columns[1]);
        v->w.I = (int) (atof(columns[2]) * 60.0 * 60.0);
        v->x.I = (int) (atof(columns[3]) * 60.0 * 60.0);

        v++;
    }

    hash_setup(g);

    fclose(fp);

    sprintf(buf, "%s/legs.csv", argv[1]);
    fp = fopen(buf, "r");
    if (fp == NULL) {
        perror(buf);
        exit(EXIT_FAILURE);
    }

    fgets(buf, sizeof(buf), fp);
    fgets(buf, sizeof(buf), fp);

    while (fgets(buf, sizeof(buf), fp)) {
        Arc *a;
        Vertex *u, *v;
        int n = parse_csv(buf, columns, MAX_COLUMNS);
        double lat1, lon1, lat2, lon2;

        u = hash_out(columns[3]);
        v = hash_out(columns[5]);

        if (u == NULL) {
            fprintf(stderr, "Cannot find fix '%s'\n", columns[3]);
            continue;
        }

        if (v == NULL) {
            fprintf(stderr, "Cannot find fix '%s'\n", columns[5]);
            continue;
        }

        lat1 = ((double) u->w.I) / (60.0 * 60.0);
        lon1 = ((double) u->x.I) / (60.0 * 60.0);
        lat2 = ((double) v->w.I) / (60.0 * 60.0);
        lon2 = ((double) v->x.I) / (60.0 * 60.0);

        assert(-90.0 <= lat1 && lat1 <= 90.0);
        assert(-180.0 <= lon1 && lon1 <= 180.0);
        assert(-90.0 <= lat2 && lat2 <= 90.0);
        assert(-180.0 <= lon2 && lon2 <= 180.0);

        gb_new_arc(u, v,
                   (int) (great_circle(RADIANS(lat1), RADIANS(lon1),
                                       RADIANS(lat2), RADIANS(lon2)) * 60.0));
        a = u->arcs;
        a->a.S = gb_save_string(columns[1]);
    }

    save_graph(g, argv[2]);

    exit(EXIT_SUCCESS);
    return 0;
}

int parse_csv(char *line, char **columns, int ncolumns) {
    int result = 0;

    while (ncolumns) {
        *columns++ = line;
        result++;

        while (*line != ',') {
            if (*line == '\r' || *line == '\n') {
                *line = '\0';
                return result;
            }

            if (*line++ == '\0')
                return result;

        }

        *line++ = '\0';
    }

    return result;
}

double
great_circle(double lat1, double lon1, double lat2, double lon2) {
        double cos_lat1 = cos(lat1), d;

	assert(-PI/2 <= lat1 && lat1 <= PI/2);
	assert(-PI <= lon1 && lon1 <= PI);
	assert(-PI/2 <= lat2 && lat2 <= PI/2);
	assert(-PI <= lon2 && lon2 <= PI);

        d = 2.0 * asin(sqrt(p2(sin((lat1 - lat2) / 2.0)) +
                            cos_lat1 * cos(lat2) *
                            p2(sin((lon2 - lon1) / 2.0))));

	assert(d >= 0.0);

	return 10800.0 * d / PI;
}

double p2(double d) {
    return d*d;
}
