#!/usr/bin/env python
#
# Copyright (c) 2006, 2010 Wayforward LLC
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307
# USA

__author__ = "Terence Way"
__email__ = "terry@wayforward.net"
__version__ = "0.1: November 12, 2010"

from datetime import date, datetime
from math import radians

TYPES = {'LOM/NDB': 'LOM',
         'LMM/NDB': 'LMM',
         'WP': 'FIX',
         'CNF': 'FIX',
         'COORDN FIX': 'FIX',
         'RPRT PT': 'FIX',
         'WAYPOINT': 'FIX',
         'RBN': 'NDB'}

AIRPORT, FIX, NAVAID, INTERSECTION = list(range(4))

TYPE_MAP = {'AIRPORT': AIRPORT,

            'FIX': FIX,
            'WP': FIX,
            'CNF': FIX,

            'BCM': NAVAID,
            'FM': NAVAID,
            'LDA': NAVAID,
            'LDA/DME': NAVAID,
            'LMM': NAVAID,
            'LOC': NAVAID,
            'LOC/DME': NAVAID,
            'LOC|SDF': NAVAID,
            'LOM': NAVAID,
            'LOM/DME': NAVAID,
            'MLS': NAVAID,
            'MLS/DME': NAVAID,
            'MLS|ISMLS': NAVAID,
            'NDB': NAVAID,
            'NDB/DME': NAVAID,
            'OM': NAVAID,
            'SDF': NAVAID,
            'TACAN': NAVAID,
            'VOR': NAVAID,
            'VOR/DME': NAVAID,
            'VORTAC': NAVAID,

            'ARTCC-BDRY': INTERSECTION,
            'AWY INT': INTERSECTION,
            'BRG INT': INTERSECTION,
            'DME FIX': INTERSECTION,
            'FIR-BDRY': INTERSECTION,
            'TRANS INT': INTERSECTION,
            'TRNG PT': INTERSECTION}

MONTHS = {'JANUARY': 1, 'FEBRUARY': 2, 'MARCH': 3,
          'APRIL': 4, 'MAY': 5, 'JUNE': 6,
          'JULY': 7, 'AUGUST': 8, 'SEPTEMBER': 9,
          'OCTOBER': 10, 'NOVEMBER': 11, 'DECEMBER': 12}

class Waypoint:
    __slots__ = ['ident', 'latitude', 'longitude', 'properties']

    def key(self):
        return (self.ident, self.latitude, self.longitude)

    def position_radians(self):
        return radians(self.latitude), radians(self.longitude)

class Navaid(Waypoint):
    __slots__ = ['type', 'magvar', 'msl_ht']

def canonical_type(type):
    return TYPES.get(type, type)

def classify_type(type):
    return TYPE_MAP.get(type)

def split_columns(line, columns):
    """Extracts columns from a line based on descriptive tuples:
        (start, end)
    or
        (start, end, type)

    >>> COLUMNS = [(0, 5), (7, 10, int)]
    >>> split_columns('test   42', COLUMNS)
    ['test', 42]

    >>> DEFS = dict(test=(0, 5), num=(7, 10, int))
    >>> split_columns('test   42', DEFS)
    {'test': 'test', 'num': 42}
    """
    if isinstance(columns, dict):
        result = {}
        for key, value in list(columns.items()):
            result[key] = split_column(line, *value)
        return result
    else:
        return [split_column(line, *t) for t in columns]

def split_column(line, start, end, type=str):
    """Extracts data from a line based on start column/end column

    >>> split_column('this is a test', 5, 7)
    'is'
    >>> split_column('this 15 a test', 5, 7, int)
    15
    """
    s = line[start:end].strip()
    if s == 'NA' and type is float: return None
    return (s or None) and type(s)

def tuple_degrees(deg, min, sec, sign):
    value = float(deg) + (float(min) + float(sec) / 60.0) / 60.0
    if sign in 'NE':
        return value
    elif sign in 'SW':
        return -value
    else:
        raise Exception('invalid latitude/longitude')

def current_cycle_date(dt):
    """Given a datetime.date, return the date of the effective FAA data
    cycle.

    Examples:
    >>> current_cycle_date(date(2010, 9, 23))
    datetime.date(2010, 9, 23)

    >>> current_cycle_date(date(2010, 9, 24))
    datetime.date(2010, 9, 23)

    >>> current_cycle_date(date(2010, 11, 17))
    datetime.date(2010, 9, 23)

    >>> current_cycle_date(date(2010, 11, 18))
    datetime.date(2010, 11, 18)
    """
    return _cycle_ordinal(dt.toordinal())

def next_cycle_date(dt):
    return _cycle_ordinal(dt.toordinal() + 56)

def _cycle_ordinal(ord):
    x, y = divmod(ord-46, 56)
    return date.fromordinal(x * 56 + 46)

def parse_navaid_datetime(datestr, timestr):
    month = int(datestr[0:2])
    day = int(datestr[2:4])
    year = int(datestr[4:6])

    hour = int(timestr[0:2])
    min = int(timestr[2:4])
    assert timestr[4] == 'Z'

    return datetime(2000 + year, month, day, hour, min, 0)

def parse_iapfix_datetime(str):
    timestr, month, day, year = str.split()
    hour = int(timestr[0:2])
    min = int(timestr[2:4])
    assert timestr[4] == 'Z'

    month = MONTHS.get(month)
    day = int(day[:-1])
    year = int(year)

    return datetime(year, month, day, hour, min, 0)

def parse_navaid_degrees(str):
    ns, deg, min, sec = str.split()
    return tuple_degrees(deg, min, sec, ns)

def parse_iapfix_degrees(str):
    deg, min, sec, ns = str.split()
    return tuple_degrees(deg, min, sec, ns)

def _test():
    import doctest, dai
    return doctest.testmod(dai)

if __name__ == '__main__':
    print(('%d/%d failures' % _test()))

